# %%
import numpy as np
from welford import Welford

# Initialize Welford object
w = Welford()

# Input data samples sequentially
w.add(np.array([0, 100]))
w.add(np.array([1, 110]))
w.add(np.array([2, 120]))

# output
print(w.mean)  # mean --> [1. 110.]
print(w.var_s)  # sample variance --> [1. 100.]
print(w.var_p)  # population variance --> [0.6666. 66.66.]

# You can add other samples after calculating variances.
w.add(np.array([3, 130]))
w.add(np.array([4, 140]))

# output with added samples
print(w.mean)  # mean --> [2. 120.]
print(w.var_s)  # sample variance --> [2.5. 250.]
print(w.var_p)  # population variance --> [2. 200.]

# You can remove samples after calculating variances.
w.remove(np.array([3, 130]))
w.remove(np.array([4, 140]))
print(w.mean)  # mean --> [1. 110.]
print(w.var_s)  # sample variance --> [1. 100.]
print(w.var_p)  # population variance --> [0.6666. 66.66.]

# You can also get the standard deviation
print(w.std_s)  # sample standard deviation --> [1. 10.]
print(w.std_p)  # population standard deviation --> [0.81649658. 8.16496581.]
# %%
import numpy as np
from welford import Welford

# Initialize Welford object with samples.
ini = np.array([[0, 100], [1, 110], [2, 120]])
w = Welford(ini)

# output
print(w.mean)  # mean --> [1. 110.]
print(w.var_s)  # sample variance --> [1. 100.]
print(w.var_p)  # population variance --> [0.66666667. 66.66666667.]

# add other samples through batch method
other_samples = np.array([[3, 130], [4, 140]])
w.add_all(other_samples)

# output with added samples
print(w.mean)  # mean --> [2. 120.]
print(w.var_s)  # sample variance --> [2.5 250.]
print(w.var_p)  # population variance --> [2. 200.]


# %%
import numpy as np
from welford import Welford

# Initialize two Welford objects
w_1 = Welford()
w_2 = Welford()

# Each object will calculate variance of each samples in parallel.
# On w_1
w_1.add(np.array([0, 100]))
w_1.add(np.array([1, 110]))
w_1.add(np.array([2, 120]))
print(w_1.var_s)  # sample variance --> [1. 100.]
print(w_1.var_p)  # population variance --> [0.66666667. 66.66666667.]

# On w_2
w_2.add(np.array([3, 130]))
w_2.add(np.array([4, 140]))
print(w_2.var_s)  # sample variance --> [0.5 50.]
print(w_2.var_p)  # sample variance --> [0.25 25.]

# You can Merge objects to get variance of WHOLE samples
w_1.merge(w_2)
print(w.var_s)  # sample variance --> [2.5. 250.]
print(w_1.var_p)  # sample variance --> [2. 200.]
# %%



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This library is a Python (Numpy) implementation of a modified Welford's algorithm,
which is online and parallel algorithm for calculating variances. Typically, Welford's algorithm
only allows for adding data points. This modification allows for removing data points.

Welford's algorithm is described in the following:

* https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Online_algorithm
* https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm

The modification for removing data points is described here:
* https://stackoverflow.com/questions/30876298/removing-a-prior-sample-while-using-welfords-method-for-computing-single-pass-v

Welford's original method is more numerically stable than the standard method as
described in the following blog.
    * Accurately computing running variance: www.johndcook.com/blog/standard_deviation

However, There has been no formal analysis on whether
the modified version of the algorithm provided here is numerically stable, but based
on the testing done in test_welford.test_remove, I have reason to believe it is.

This library is inspired by the jvf's implementation, which is implemented
without using numpy library. In particular, this implementation is a fork
of the implementation by a-mitani,
    * Implementation done by jvf: github.com/jvf/welford
    * Implementation done by a-mitani: github.com/a-mitani/welford
"""
from __future__ import annotations

import numpy as np
import numpy.typing as npt
from typing import Optional


class Welford:
    """Accumulator object for Welford's online and parallel variance algorithm which provides the ability to remove data points.

    Attributes:
        count (int): The number of accumulated samples.
        mean (np.array(D,)): Mean of the accumulated samples.
        var_s (np.array(D,)): Sample variance of the accumulated samples.
        var_p (np.array(D,)): Population variance of the accumulated samples.
        std_s (np.array(D,)): Sample standard deviation of the accumulated samples.
        std_p (np.array(D,)): Population standard deviation of the accumulated samples.
    """

    def __init__(self, elements: Optional[npt.NDArray[npt.NDArray]] = None):
        """
        Initialize with an optional data. 
        For calculation efficiency, Welford's method is not used in the initialization process.

        Args:
            elements: The data points to initialize with.

        """
        self.__count: int = 0
        self.__shape: Optional[tuple]
        self.__m: Optional[npt.NDArray]
        self.__s: Optional[npt.NDArray]
        self.__m_old: Optional[npt.NDArray]
        self.__s_old: Optional[npt.NDArray]
        self.__count_old: Optional[int]

        # Initialize instance attributes
        if elements is None:
            self.__shape = None
            # Current attribute values
            self.__count = 0
            self.__m = None
            self.__s = None
            # Previous attribute values for rolling back
            self.__count_old = None
            self.__m_old = None
            self.__s_old = None
        else:
            self.__shape = elements[0].shape
            # Current attribute values
            self.__count = elements.shape[0]
            self.__m = np.array(np.mean(elements, axis=0))
            self.__s = np.array(np.var(elements, axis=0, ddof=0) * elements.shape[0])
            # Previous attribute values for rolling back
            self.__count_old = None
            self.__init_old_with_nan()

    @property
    def count(self) -> Optional[int]:
        """Get the number of accumulated samples."""
        return self.__count

    @property
    def mean(self) -> Optional[npt.NDArray]:
        """Get the mean."""
        return self.__m

    @property
    def var_s(self) -> Optional[npt.NDArray]:
        """Get the sample variance."""
        return self.__get_var_with_ddof(ddof=1)

    @property
    def var_p(self) -> Optional[npt.NDArray]:
        """Get the population variance."""
        return self.__get_var_with_ddof(ddof=0)

    @property
    def std_s(self) -> Optional[npt.NDArray]:
        """Get the sample standard deviation."""
        if self.var_s is None:
            return None

        return np.sqrt(self.var_s, dtype=np.longdouble)

    @property
    def std_p(self) -> Optional[npt.NDArray]:
        """Get the population standard deviation."""
        if self.var_p is None:
            return None

        return np.sqrt(self.var_p, dtype=np.longdouble)

    def add(self, element: npt.NDArray, backup_flg: bool = True) -> None:
        """Add one data point.

        Args:
            element: The data point to add.
            backup_flg: If True, backup current state for rolling back.

        """

        # Initialize if not done so.
        if self.__shape is None:
            self.__shape = element.shape
            self.__m = np.zeros(element.shape)
            self.__s = np.zeros(element.shape)
            self.__init_old_with_nan()
            self.__count = 0
        else:
            # Argument check if already initialized
            assert element.shape == self.__shape

        if backup_flg:
            self.__backup_attrs()

        # Welford's algorithm
        self.__count += 1
        delta = element - self.__m
        self.__m += delta / self.__count
        self.__s += delta * (element - self.__m)

    def add_all(self, elements: npt.NDArray[npt.NDArray], backup_flg: bool = True) -> None:
        """Add multiple data points.

        Args:
            elements: The data points to add.
            backup_flg: If True, backup current state for rolling back.

        """
        if backup_flg:
            self.__backup_attrs()

        for elem in elements:
            self.add(elem, backup_flg=False)

    def remove(self, element: npt.NDArray, backup_flg: bool = True) -> None:
        """Remove one data point.

        Using a method derived from the original Welford's algorithm,
        we can remove a data point. Note however, there has been
        no analysis on whether this is numerically stable. See
        https://stackoverflow.com/questions/30876298/removing-a-prior-sample-while-using-welfords-method-for-computing-single-pass-v
        for more information.

        """

        if self.__shape is None:
            return
        else:
            assert element.shape == self.__shape

        if backup_flg:
            self.__backup_attrs()

        # The reverse of Welford's algorithm.
        old_m = np.copy(self.__m)
        self.__m -= (element - self.__m) / (self.count - 1)
        self.__s -= (element - self.__m) * (element - old_m)
        self.__count -= 1

    def remove_all(self, elements: npt.NDArray[npt.NDArray], backup_flg: bool = True) -> None:
        """Remove multiple data points.

        Args:
            elements: The data points to remove.
            backup_flg: If True, backup current state for rolling back.

        """
        if backup_flg:
            self.__backup_attrs()

        for elem in elements:
            self.remove(elem, backup_flg=False)

    def rollback(self) -> None:
        """
        Rollback to a prior state that has been saved.
        In order for a state to be saved, you must call
        pass backup_flag = True to either the method merge,
        add, remove, add_all, or remove_all."""

        if self.__shape is None:
            return

        self.__count = self.__count_old
        self.__m[...] = self.__m_old[...]
        self.__s[...] = self.__s_old[...]

    def merge(self, other: Welford, backup_flg: bool = True) -> None:
        """Merge this accumulator with another one.

        Args:
            other: The other accumulator to merge with.
            backup_flg: If True, backup current state for rolling back.

        """

        if other.__shape is None:
            pass
        elif self.__shape is None:
            if backup_flg:
                self.__backup_attrs()

            self.__count = other.__count
            self.__m = other.__m
            self.__s = other.__s
        else:
            if backup_flg:
                self.__backup_attrs()

            count = self.__count + other.__count
            delta = self.__m - other.__m
            delta2 = delta * delta
            m = (self.__count * self.__m + other.__count * other.__m) / count
            s = self.__s + other.__s + delta2 * (self.__count * other.__count) / count

            self.__count = count
            self.__m = m
            self.__s = s

    def __get_var_with_ddof(self, ddof: int) -> Optional[npt.NDArray]:
        """
        Get the variance with a given delta degrees of freedom (ddof).
        If there are fewer sample accumulated than the ddof, return a matrix
        filled with NaN values of the appropriate shape.

        Args:
            ddof: The delta degrees of freedom.

        """
        if self.__count <= 0:
            return None
        min_count = ddof
        if self.__count <= min_count:
            return np.full(self.__shape, np.nan)
        else:
            return self.__s / (self.__count - ddof)

    def __backup_attrs(self) -> None:
        """
        Backup current values for mean and variance. Used for when you want to rollback changes
        made through adding or removing samples.
        """
        if self.__shape is None:
            pass
        else:
            self.__count_old = self.__count
            self.__m_old[...] = self.__m[...]
            self.__s_old[...] = self.__s[...]

    def __init_old_with_nan(self) -> None:
        """
        Initialize old values with NaN.
        """
        self.__m_old = np.empty(self.__shape)
        self.__m_old[...] = np.nan
        self.__s_old = np.empty(self.__shape)
        self.__s_old[...] = np.nan



import numpy as np
from welford import Welford


def test_init():
    a = np.array([[0]])
    w = Welford(a)
    assert w.count == 1
    assert np.allclose(w.mean, np.array([0]))
    assert np.alltrue(np.isnan(w.var_s))
    assert np.allclose(w.var_p, np.array([0]))

    a = np.array([[0], [1]])
    w = Welford(a)
    assert w.count == 2
    assert np.allclose(w.mean, np.array([0.5]))
    assert np.allclose(w.var_s, np.array([0.5]))
    assert np.allclose(w.var_p, np.array([0.25]))

    a = np.array([[0, 100], [1, 110], [2, 120], [3, 130], [4, 140]])
    w = Welford(a)
    assert w.count == 5
    assert np.allclose(w.mean, np.array([2, 120]))
    assert np.allclose(w.var_s, np.array([2.5, 250]))
    assert np.allclose(w.var_p, np.array([2, 200]))

    a = np.arange(60).reshape(3, 4, 5)
    w = Welford(a)
    assert w.mean.shape == w.var_s.shape == w.var_p.shape == (4, 5)
    assert np.allclose(w.mean, np.mean(a, axis=0))
    assert np.allclose(w.var_s, np.var(a, axis=0, ddof=1))
    assert np.allclose(w.var_p, np.var(a, axis=0, ddof=0))


def test_add():
    w = Welford()
    w.add(np.array([0, 100]))
    assert np.allclose(w.mean, np.array([0, 100]))
    assert np.alltrue(np.isnan(w.var_s))
    assert np.allclose(w.var_p, np.array([0, 0]))

    w.add(np.array([1, 110]))
    assert np.allclose(w.mean, np.array([0.5, 105]))
    assert np.allclose(w.var_s, np.array([0.5, 50]))
    assert np.allclose(w.var_p, np.array([0.25, 25]))

    w = Welford()
    w.add(np.array([[0, 100, 1000], [2, 220, 2200]]))
    w.add(np.array([[1, 110, 1100], [2, 220, 2200]]))
    assert np.allclose(w.mean, np.array([[0.5, 105, 1050], [2, 220, 2200]]))
    assert np.allclose(w.var_s, np.array([[0.5, 50, 5000], [0, 0, 0]]))
    assert np.allclose(w.var_p, np.array([[0.25, 25, 2500], [0, 0, 0]]))


def test_add_all():
    w = Welford()
    a = np.array([[0, 100], [1, 110], [2, 120], [3, 130], [4, 140]])
    w.add_all(a)
    assert w.count == 5
    assert np.allclose(w.mean, np.array([2, 120]))
    assert np.allclose(w.var_s, np.array([2.5, 250]))
    assert np.allclose(w.var_p, np.array([2, 200]))

    w = Welford()
    a = np.array([[0, 100]])
    w.add_all(a)
    assert w.count == 1
    assert np.allclose(w.mean, np.array([0, 100]))
    assert np.alltrue(np.isnan(w.var_s))
    assert np.allclose(w.var_p, np.array([0, 0]))

    a = np.array([[1, 110], [2, 120], [3, 130], [4, 140]])
    w.add_all(a)
    assert w.count == 5
    assert np.allclose(w.mean, np.array([2, 120]))
    assert np.allclose(w.var_s, np.array([2.5, 250]))
    assert np.allclose(w.var_p, np.array([2, 200]))

    w = Welford()
    a = np.array([[[0, 100, 1000], [2, 220, 2200]], [[1, 110, 1100], [2, 220, 2200]]])
    w.add_all(a)
    assert w.count == 2
    assert np.allclose(w.mean, np.array([[0.5, 105, 1050], [2, 220, 2200]]))
    assert np.allclose(w.var_s, np.array([[0.5, 50, 5000], [0, 0, 0]]))
    assert np.allclose(w.var_p, np.array([[0.25, 25, 2500], [0, 0, 0]]))


def test_rollback():
    a = np.array([[0, 100]])
    w = Welford(a)

    a = np.array([1, 110])
    w.add(a)
    assert np.allclose(w.mean, np.array([0.5, 105]))
    assert np.allclose(w.var_s, np.array([0.5, 50]))
    assert np.allclose(w.var_p, np.array([0.25, 25]))

    w.rollback()
    a = np.array([2, 120])
    w.add(a)
    assert w.count == 2
    assert np.allclose(w.mean, np.array([1, 110]))
    assert np.allclose(w.var_s, np.array([2, 200]))
    assert np.allclose(w.var_p, np.array([1, 100]))

    a = np.array([[2, 120], [3, 130]])
    w.add_all(a)
    w.rollback()
    assert w.count == 2
    assert np.allclose(w.mean, np.array([1, 110]))
    assert np.allclose(w.var_s, np.array([2, 200]))
    assert np.allclose(w.var_p, np.array([1, 100]))

    w = Welford()
    w.add(np.array([[0, 100, 1000], [2, 220, 2200]]))
    w.add(np.array([[1, 110, 1100], [2, 220, 2200]]))
    w.rollback()
    w.add(np.array([[2, 120, 1200], [2, 220, 2200]]))
    assert np.allclose(w.mean, np.array([[1.0, 110, 1100], [2, 220, 2200]]))
    assert np.allclose(w.var_s, np.array([[2.0, 200, 20000], [0, 0, 0]]))
    assert np.allclose(w.var_p, np.array([[1, 100, 10000], [0, 0, 0]]))

def test_merge():
    a = np.array([[0, 100], [1, 110], [2, 120], [3, 130], [4, 140]])
    wa = Welford(a)
    b = np.array([[5, 150], [6, 160], [7, 170]])
    wb = Welford(b)
    wa.merge(wb)
    assert wa.count == 8
    assert np.allclose(wa.mean, np.array([3.5, 135]))
    assert np.allclose(wa.var_s, np.array([6, 600]))
    assert np.allclose(wa.var_p, np.array([5.25, 525]))

    wa.rollback()
    c = np.array([[5, 150], [6, 160], [7, 170], [8, 180], [9, 190], [10, 200]])
    wc = Welford(c)
    wa.merge(wc)
    assert wa.count == 11
    assert np.allclose(wa.mean, np.array([5, 150]))
    assert np.allclose(wa.var_s, np.array([11, 1100]))
    assert np.allclose(wa.var_p, np.array([10, 1000]))

    a = np.array([[0, 100]])
    wa = Welford(a)
    b = np.array([[1, 110]])
    wb = Welford(b)
    wa.merge(wb)
    assert wa.count == 2
    assert np.allclose(wa.mean, np.array([0.5, 105]))
    assert np.allclose(wa.var_s, np.array([0.5, 50]))
    assert np.allclose(wa.var_p, np.array([0.25, 25]))

    wa = Welford(np.array([[[0, 100, 1000], [2, 220, 2200]]]))
    wb = Welford(np.array([[[1, 110, 1100], [2, 220, 2200]]]))
    wa.merge(wb)
    assert wa.count == 2
    assert np.allclose(wa.mean, np.array([[0.5, 105, 1050], [2, 220, 2200]]))
    assert np.allclose(wa.var_s, np.array([[0.5, 50, 5000], [0, 0, 0]]))
    assert np.allclose(wa.var_p, np.array([[0.25, 25, 2500], [0, 0, 0]]))


def test_merge_empty_other():
    a = np.array([[0, 100], [1, 110], [2, 120], [3, 130], [4, 140]])

    wa = Welford(a)
    wb = Welford()
    wa.merge(wb)
    assert wa.count == 5
    assert np.allclose(wa.mean, np.array([2, 120]))
    assert np.allclose(wa.var_s, np.array([2.5, 250]))
    assert np.allclose(wa.var_p, np.array([2, 200]))
    # assert

def test_merge_empty_self():
    b = np.array([[0, 100], [1, 110], [2, 120], [3, 130], [4, 140]])

    wa = Welford()
    wb = Welford(b)
    wa.merge(wb)
    assert wa.count == 5
    assert np.allclose(wa.mean, np.array([2, 120]))
    assert np.allclose(wa.var_s, np.array([2.5, 250]))
    assert np.allclose(wa.var_p, np.array([2, 200]))


def test_merge_both_empty():
    wa = Welford()
    wb = Welford()
    wa.merge(wb)
    assert wa.count == wb.count == 0
    assert not wa.mean
    assert not wa.var_s
    assert not wa.var_p


def test_add_many():
    w = Welford()
    running_sample = []
    N = 1000

    for i in range(0, N):
        random_number = np.array([np.random.rand()])
        running_sample.append(random_number)
        w.add(random_number)

        if i == 0:
            continue

        std = np.std(running_sample, ddof=1, dtype=np.longdouble)
        mean = np.mean(running_sample, dtype=np.longdouble)

        assert np.isclose(std, w.std_s)
        assert np.isclose(mean, w.mean)


def test_remove():
    N = 100000
    random_sample = np.random.rand(N)
    w = Welford(random_sample)

    for i in range(0, N - 2):
        w.remove(random_sample[0])
        random_sample = random_sample[1:]

        std = np.std(random_sample, ddof=1, dtype=np.longdouble)
        mean = np.mean(random_sample, dtype=np.longdouble)

        assert np.isclose(std, w.std_s)
        assert np.isclose(mean, w.mean)



from setuptools import setup, find_packages
import pypandoc
try:

    long_description = pypandoc.convert_file('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()

setup(
    name='welford-remove',
    version=0.1,
    description="Python (numpy) implementation of Welford's algorithm with the ability to remove data points.",
    author="Robert Gold",
    author_email="18goldr@gmail.com",
    url="https://github.com/18goldr/welford-with-remove",
    license='MIT',
    keywords=["statistics", "online", "welford"],
    install_requires=["numpy"],
    long_description=long_description,
    packages=find_packages(),

)
